@@ -20,14 +20,37 @@
 
 #include <string.h>
 
+#include <gdk/gdkkeysyms.h>
+
 #include "terminal-accels.h"
 #include "terminal-app.h"
 #include "terminal-intl.h"
 #include "terminal-profile.h"
 #include "terminal-util.h"
+#include "terminal-accels.h"
 
+#ifdef DEBUG_ACCELS
+#define D(x) x
+#else
 #define D(x)
+#endif
 
+/* NOTES
+ *
+ * There are two sources of keybindings changes, from GConf and from
+ * the accel map (happens with in-place menu editing).
+ *
+ * When a keybinding gconf key changes, we propagate that into the
+ * accel map.
+ * When the accel map changes, we queue a sync to gconf.
+ *
+ * To avoid infinite loops, we short-circuit in both directions
+ * if the value is unchanged from last known.
+ *
+ * In the keybinding editor, when editing or clearing an accel, we write
+ * the change directly to gconf and rely on the gconf callback to
+ * actually apply the change to the accel map.
+ */
 
 #define ACCEL_PATH_ROOT "<Actions>/Main/"
 #define ACCEL_PATH_NEW_TAB              ACCEL_PATH_ROOT "FileNewTab"
@@ -51,6 +74,7 @@
 #define ACCEL_PATH_MOVE_TAB_LEFT        ACCEL_PATH_ROOT "TabsMoveLeft"
 #define ACCEL_PATH_MOVE_TAB_RIGHT       ACCEL_PATH_ROOT "TabsMoveRight"
 #define ACCEL_PATH_DETACH_TAB           ACCEL_PATH_ROOT "TabsDetach"
+#define ACCEL_PATH_SWITCH_TAB_PREFIX    ACCEL_PATH_ROOT "TabsSwitch"
 
 #define KEY_CLOSE_TAB           CONF_KEYS_PREFIX "/close_tab"
 #define KEY_CLOSE_WINDOW        CONF_KEYS_PREFIX "/close_window"
@@ -73,6 +97,7 @@
 #define KEY_ZOOM_IN             CONF_KEYS_PREFIX "/zoom_in"
 #define KEY_ZOOM_NORMAL         CONF_KEYS_PREFIX "/zoom_normal"
 #define KEY_ZOOM_OUT            CONF_KEYS_PREFIX "/zoom_out"
+#define KEY_SWITCH_TAB_PREFIX   CONF_KEYS_PREFIX "/switch_to_tab_"
 
 typedef struct
 {
@@ -141,7 +166,7 @@ static KeyEntry terminal_entries[] =
     KEY_RESET_AND_CLEAR, ACCEL_PATH_RESET_AND_CLEAR, 0, 0, NULL, FALSE, TRUE },
 };
 
-static KeyEntry go_entries[] =
+static KeyEntry tabs_entries[] =
 {
   { N_("Switch to Previous Tab"),
     KEY_PREV_TAB, ACCEL_PATH_PREV_TAB, 0, 0, NULL, FALSE, TRUE },
@@ -155,6 +180,47 @@ static KeyEntry go_entries[] =
     KEY_DETACH_TAB, ACCEL_PATH_DETACH_TAB, 0, 0, NULL, FALSE, TRUE },
 };
 
+#define UN_(x) (x)
+
+static KeyEntry jump_tab_entries[] = {
+  { UN_("Switch to Tab 1"),
+    KEY_SWITCH_TAB_PREFIX "1",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "1", GDK_1, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 2"),
+    KEY_SWITCH_TAB_PREFIX "2",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "2", GDK_2, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 3"),
+    KEY_SWITCH_TAB_PREFIX "3",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "3", GDK_3, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 4"),
+    KEY_SWITCH_TAB_PREFIX "4",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "4", GDK_4, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 5"),
+    KEY_SWITCH_TAB_PREFIX "5",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "5", GDK_5, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 6"),
+    KEY_SWITCH_TAB_PREFIX "6",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "6", GDK_6, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 7"),
+    KEY_SWITCH_TAB_PREFIX "7",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "7", GDK_7, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 8"),
+    KEY_SWITCH_TAB_PREFIX "8",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "8", GDK_8, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 9"),
+    KEY_SWITCH_TAB_PREFIX "9",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "9", GDK_9, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 10"),
+    KEY_SWITCH_TAB_PREFIX "10",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "10", GDK_0, GDK_MOD1_MASK, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 11"),
+    KEY_SWITCH_TAB_PREFIX "11",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "11", 0, 0, NULL, FALSE, TRUE },
+  { UN_("Switch to Tab 12"),
+    KEY_SWITCH_TAB_PREFIX "12",
+    ACCEL_PATH_SWITCH_TAB_PREFIX "12", 0, 0, NULL, FALSE, TRUE }
+};
+
 static KeyEntry help_entries[] = {
   { N_("Contents"), KEY_HELP, ACCEL_PATH_HELP, 0, 0, NULL, FALSE, TRUE }
 };
@@ -165,8 +231,9 @@ static KeyEntryList all_entries[] =
   { edit_entries, G_N_ELEMENTS (edit_entries), N_("Edit") },
   { view_entries, G_N_ELEMENTS (view_entries), N_("View") },
   { terminal_entries, G_N_ELEMENTS (terminal_entries), N_("Terminal") },
-  { go_entries, G_N_ELEMENTS (go_entries), N_("Go") },
-  { help_entries, G_N_ELEMENTS (help_entries), N_("Help") }
+  { tabs_entries, G_N_ELEMENTS (tabs_entries), N_("Tabs") },
+  { help_entries, G_N_ELEMENTS (help_entries), N_("Help") },
+  { jump_tab_entries, G_N_ELEMENTS (jump_tab_entries), NULL }
 };
 
 enum
@@ -176,15 +243,6 @@ enum
   N_COLUMNS
 };
 
-/*
- * This is kind of annoying. We have two sources of keybinding change;
- * GConf and GtkAccelMap. GtkAccelMap will change if the user uses
- * the magic in-place editing mess. If accel map changes, we propagate
- * into GConf. If GConf changes we propagate into accel map.
- * To avoid infinite loop hell, we short-circuit in both directions
- * if the value is unchanged from last known.
- */
-
 static void keys_change_notify (GConfClient *client,
                                 guint        cnxn_id,
                                 GConfEntry  *entry,
@@ -207,10 +265,11 @@ static gboolean binding_from_value  (GConfValue       *value,
 static gboolean sync_idle_cb (gpointer data);
 
 static guint sync_idle_id = 0;
-static GtkAccelGroup * /* accel_group_i_need_because_gtk_accel_api_sucks */ hack_group = NULL;
+static GtkAccelGroup *notification_group = NULL;
 /* never set gconf keys in response to receiving a gconf notify. */
 static int inside_gconf_notify = 0;
 static GtkWidget *edit_keys_dialog = NULL;
+static GtkTreeStore *edit_keys_store = NULL;
 static guint gconf_notify_id;
 static GHashTable *gconf_key_to_entry;
 
@@ -262,7 +321,7 @@ terminal_accels_init (void)
 
   gconf_key_to_entry = g_hash_table_new (g_str_hash, g_str_equal);
 
-  hack_group = gtk_accel_group_new ();
+  notification_group = gtk_accel_group_new ();
 
   for (i = 0; i < G_N_ELEMENTS (all_entries); ++i)
     {
@@ -281,7 +340,7 @@ terminal_accels_init (void)
 	  g_closure_ref (key_entry->closure);
 	  g_closure_sink (key_entry->closure);
 	  
-	  gtk_accel_group_connect_by_path (hack_group,
+	  gtk_accel_group_connect_by_path (notification_group,
 					   I_(key_entry->accel_path),
 					   key_entry->closure);
 
@@ -289,7 +348,7 @@ terminal_accels_init (void)
 	}
     }
   
-  g_signal_connect (hack_group, "accel_changed",
+  g_signal_connect (notification_group, "accel-changed",
                     G_CALLBACK (accel_changed_callback), NULL);
 }
 
@@ -357,6 +416,7 @@ keys_change_notify (GConfClient *client,
               gconf_value_get_string (val)));
 
   key_entry = g_hash_table_lookup (gconf_key_to_entry, key_from_gconf_key (gconf_entry_get_key (entry)));
+  D (if (!key_entry) g_print ("  WARNING: KeyEntry for changed key not found, bailing out\n"));
   if (!key_entry)
     return; /* shouldn't really happen, but let's be safe */
 
@@ -387,10 +447,21 @@ keys_change_notify (GConfClient *client,
               key_entry->accel_path,
               binding_name (keyval, mask))); /* memleak */
   inside_gconf_notify += 1;
+  /* Note that this may return FALSE, e.g. when the entry was already set correctly. */
   gtk_accel_map_change_entry (key_entry->accel_path,
                               keyval, mask,
                               TRUE);
   inside_gconf_notify -= 1;
+
+  /* This seems necessary to update the tree model, since sometimes the
+   * notification on the notification_group seems not to be emitted correctly.
+   * Without this change, when trying to set an accel to e.g. Alt-T (while the main
+   * menu in the terminal windows is _Terminal with Alt-T mnemonic) only displays
+   * the accel change after a re-expose of the row.
+   * FIXME: Find out *why* the accel-changed signal is wrong here!
+   */
+  if (edit_keys_store)
+    gtk_tree_model_foreach (GTK_TREE_MODEL (edit_keys_store), update_model_foreach, key_entry);
 }
 
 static void
@@ -612,6 +683,7 @@ treeview_accel_changed_cb (GtkAccelGroup  *accel_group,
                            GClosure *accel_closure,
                            GtkTreeModel *model)
 {
+g_print("TREEVIEW ACCEL-CHANGED keyval %s mask %x\n", gdk_keyval_name (keyval), modifier);
   gtk_tree_model_foreach (model, update_model_foreach, accel_closure->data);
 }
 
@@ -651,7 +723,7 @@ accel_edited_callback (GtkCellRendererAccel *cell,
     return;
 
   /* Check if we already have an entry using this accel */
-  entries = gtk_accel_group_query (hack_group, keyval, mask, &n_entries);
+  entries = gtk_accel_group_query (notification_group, keyval, mask, &n_entries);
   if (n_entries > 0)
     {
       if (entries[0].accel_path_quark != g_quark_from_string (ke->accel_path))
@@ -683,9 +755,17 @@ accel_edited_callback (GtkCellRendererAccel *cell,
 
   str = binding_name (keyval, mask);
 
-  D (g_print ("Edited keyval %s, setting gconf to %s\n",
+  D (g_print ("Edited path %s keyval %s, setting gconf to %s\n",
+              ke->accel_path,
               gdk_keyval_name (keyval) ? gdk_keyval_name (keyval) : "null",
               str));
+  D({
+    GtkAccelKey old_key;
+    if (gtk_accel_map_lookup_entry (ke->accel_path, &old_key)) {
+      g_print ("  Old entry of path %s is keyval %s mask %x\n", ke->accel_path, gdk_keyval_name (old_key.accel_key), old_key.accel_mods);
+    } else
+      g_print ("  Failed to look up the old entry of path %s\n", ke->accel_path);
+    })
 
   conf = gconf_client_get_default ();
   gconf_client_set_string (conf,
@@ -746,10 +826,36 @@ static void
 edit_keys_dialog_destroy_cb (GtkWidget *widget,
                              gpointer user_data)
 {
-  g_signal_handlers_disconnect_by_func (hack_group, G_CALLBACK (treeview_accel_changed_cb), user_data);
+  g_signal_handlers_disconnect_by_func (notification_group, G_CALLBACK (treeview_accel_changed_cb), user_data);
   edit_keys_dialog = NULL;
+  edit_keys_store = NULL;
 }
 
+static void
+edit_keys_dialog_response_cb (GtkWidget *editor,
+                              int response,
+                              gpointer use_data)
+{  
+  if (response == GTK_RESPONSE_HELP)
+    {
+      terminal_util_show_help ("gnome-terminal-shortcuts", GTK_WINDOW (editor));
+      return;
+    }
+    
+  gtk_widget_destroy (editor);
+}
+
+#ifdef DEBUG_ACCELS 
+static void
+row_changed (GtkTreeModel *tree_model,
+             GtkTreePath  *path,
+             GtkTreeIter  *iter,
+             gpointer      user_data)
+{
+  g_print ("ROW-CHANGED [%s]\n", gtk_tree_path_to_string (path));
+}
+#endif
+
 void
 terminal_edit_keys_dialog_show (GtkWindow *transient_parent)
 {
@@ -806,8 +912,13 @@ terminal_edit_keys_dialog_show (GtkWindow *transient_parent)
 
   /* Add the data */
 
-  tree = gtk_tree_store_new (N_COLUMNS, G_TYPE_STRING, G_TYPE_POINTER);
-  for (i = 0; i < G_N_ELEMENTS (all_entries); ++i)
+  tree = edit_keys_store = gtk_tree_store_new (N_COLUMNS, G_TYPE_STRING, G_TYPE_POINTER);
+#ifdef DEBUG_ACCELS
+  g_signal_connect (tree, "row-changed", G_CALLBACK (row_changed), NULL);
+#endif
+
+  /* -1 because we can't add the Tabs entries on gnome-2-24 b/c of string freeze */
+  for (i = 0; i < G_N_ELEMENTS (all_entries) - 1; ++i)
     {
       GtkTreeIter parent_iter;
       int j;
@@ -840,14 +951,14 @@ terminal_edit_keys_dialog_show (GtkWindow *transient_parent)
 
   gtk_tree_view_expand_all (GTK_TREE_VIEW (tree_view));
 
-  g_signal_connect (hack_group, "accel-changed",
+  g_signal_connect (notification_group, "accel-changed",
                     G_CALLBACK (treeview_accel_changed_cb), tree);
 
   edit_keys_dialog = dialog;
   g_signal_connect (dialog, "destroy",
                     G_CALLBACK (edit_keys_dialog_destroy_cb), tree);
   g_signal_connect (dialog, "response",
-                    G_CALLBACK (gtk_widget_destroy),
+                    G_CALLBACK (edit_keys_dialog_response_cb),
                     NULL);
   gtk_window_set_default_size (GTK_WINDOW (dialog), -1, 350);
 
